package lezione6.scommesse.client;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.MulticastSocket;
import java.net.Socket;
import java.net.UnknownHostException;
import java.security.KeyFactory;
import java.security.PublicKey;
import java.security.spec.X509EncodedKeySpec;
import java.util.Arrays;

import javax.crypto.Cipher;

/**
 * Questo è un esempio di un'applicazione che garantisce la segretezza della
 * comunicazione, in quanto cifra i dati da inviare verso il server con la
 * chiave pubblica comunicat dal server stesso. Solo il server sarà capace di
 * decifrare questi dati, in quanto è l'unico che dispone della chiave privata
 * accoppiata alla chiave pubblica.
 * 
 * @author Raffaele Giordanelli
 * 
 */
class BetClient {

	int serverPort;
	int myPort;

	InetAddress groupAddress;
	InetAddress serverAddress;

	Socket s;

	// Cryptography: begin

	// Socket per richiedere la chiave pubblica del bet server
	Socket gks;
	// Porta del server e socket a cui richiedere la chiave pubblica
	int getKeyServerPort;
	// Cipher per cifrare i dati da inviare al server
	Cipher cipher;
	// Cipher per decifrare i dati da inviati dal server
	Cipher decipher;

	// Cryptography: end

	public BetClient(InetAddress gAddress, InetAddress server, int sPort, int mPort, int gkServerPort) {

		groupAddress = gAddress;
		serverAddress = server;
		serverPort = sPort;
		myPort = mPort;
		getKeyServerPort = gkServerPort;
		try {
			s = new Socket(serverAddress, serverPort);

			// Cryptography: begin
			// Inizializzazione del socket per richiedere la chiave pubblica del
			// bet server
			gks = new Socket(serverAddress, getKeyServerPort);
			// Cryptography: end

		} catch (IOException ioe) {
			System.out.println(ioe);
		}
	}

	/**
	 * Legge la chiave pubblica inviata dal server, e tramite questa inizializza
	 * i due cifrari RSA in modalità cifratura e decifratura
	 */
	public void makeCipher() {
		try {
			// Cryptography: begin
			// Recupero gli Stream di comunicazione
			InputStream inGetKey = gks.getInputStream();
			OutputStream outGetKey = gks.getOutputStream();

			// Richiedo la chiave publica del Bet Server
			outGetKey.write(1);

			// Salvo la chiave inviata dal server in publicKeyBytes
			byte[] publicKeyBytes = new byte[1000];
			int numByte = inGetKey.read(publicKeyBytes);
			publicKeyBytes = Arrays.copyOf(publicKeyBytes, numByte);

			// Inizializza convertitore da X.509 a chiave pubblica
			X509EncodedKeySpec ks = new X509EncodedKeySpec(publicKeyBytes);
			// Inizializza un KeyFactory per ricreare la chiave usando RSA
			KeyFactory kf = KeyFactory.getInstance("RSA");
			// Crea una chiave pubblica usando generatePublic di KeyFactory in
			// base la chiave decodificata da ks
			PublicKey publicKey = kf.generatePublic(ks);

			// Inizializzo un cifrario che usa come algoritmo RSA, come modalità
			// ECB e come padding PKCS1
			cipher = Cipher.getInstance("RSA/ECB/PKCS1Padding");

			// Lo inizializzo dicendo modalità di codifica e chiave pubblica da
			// usare
			cipher.init(Cipher.ENCRYPT_MODE, publicKey);

			// Inizializzo un cifrario che usa come algoritmo RSA, come modalità
			// ECB e come padding PKCS1
			decipher = Cipher.getInstance("RSA/ECB/PKCS1Padding");

			// Lo inizializzo dicendo modalità di codifica e chiave pubblica da
			// usare
			decipher.init(Cipher.DECRYPT_MODE, publicKey);

			// Cryptography: end
		} catch (Exception e) {
			System.out.println(e);
		}

	}

	/**
	 * Invia la scommessa al server
	 * 
	 * @return true, se la scommessa è stata accettata, false altrimenti
	 */
	public boolean placeBet(int nCavallo, long puntata) {

		// Cryptography: begin2
		// Creo il Cipher per criptare i dati inviati
		makeCipher();
		// Cryptography: end

		String e = "";
		try {
			BufferedReader in = new BufferedReader(new InputStreamReader(s.getInputStream()));
			OutputStream out = s.getOutputStream();

			byte[] bet = cipher.doFinal((nCavallo + " " + puntata).getBytes());
			out.write(bet);

			e = in.readLine();
		} catch (Exception err) {
			System.out.println(err);
		}
		return e.equals("Scommessa accettata.");
	}

	/**
	 * Riceve l'elenco dei vincitori inviato dal server in multicast a tutti i
	 * client
	 */
	public void riceviElenco() {
		try {
			MulticastSocket socket = new MulticastSocket(myPort);
			socket.joinGroup(groupAddress);

			byte[] buf = new byte[128];
			DatagramPacket packet = new DatagramPacket(buf, buf.length);
			socket.receive(packet);

			// Cryptography: begin
			String elenco = new String(decipher.doFinal(packet.getData()));
			// Cryptography: end

			System.out.println("Elenco vincitori: ");
			System.out.println(elenco);

		} catch (IOException ioe) {
			System.out.println(ioe);
		} catch (Exception e) {
			System.out.println(e);
		}
	}

	public static void main(String[] args) {

		int serverPort = 8001;
		int myPort = 8002;

		// Cryptography: begin
		int getKeyServerPort = 8443;
		// Cryptography: end

		try {
			InetAddress group = InetAddress.getByName("230.0.0.1");
			InetAddress server = InetAddress.getByName("127.0.0.1");

			BetClient client = new BetClient(group, server, serverPort, myPort, getKeyServerPort);

			int nCavallo = ((int) (Math.random() * 12)) + 1;
			if (client.placeBet(nCavallo, 15))
				client.riceviElenco();

		} catch (UnknownHostException uhe) {
			System.out.println(uhe);
		}

	}// main

}// BetClient