package lezione6.scommesse.server;

/*Corse dei Cavalli*/

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.MulticastSocket;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketTimeoutException;
import java.net.UnknownHostException;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.util.Arrays;
import java.util.Calendar;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.ListIterator;

import javax.crypto.Cipher;

import lezione6.scommesse.Scommessa;

/**
 * Questo è un esempio di un'applicazione che garantisce autenticazione del
 * mittente. Il server genera una coppia di chiavi e comunica la chiave pubblica
 * a chi ne fa richiesta. In seguito ogni comunicazione in uscita sarà cifrata
 * con la chiave privata. I client, che decifrano i messaggi con la chiave
 * pubblica sapranno che può essere stato solo questo server ad inviare i dati,
 * perché è l'unica entità a disposizione della chiave privata originale.
 * 
 * @author Raffaele Giordanelli
 * 
 */
class BetServer {

	private HashMap<Integer, Scommessa> scommesse;
	private Calendar limite;
	private BetAccepter accepter;
	private BetDenier denier;
	private int port;

	// Cryptography: begin
	// Thread per la gestione dell'invio della chiave pubblica
	private SendKey sendKey;
	// Cipher per la decifrazione dei dati inviati dallo scommettitore
	private Cipher decipher;
	// Cipher per la cifrazione dei dati da inviare allo scommettitore
	private Cipher cipher;
	// Coppia di chiavi pubblica e privata
	KeyPair kp;

	// Cryptography: end

	public BetServer(int port, int getKeyPort, Calendar deadline) {

		// Cryptography: begin
		// Avvio il server per l'invio della chiave pubblica
		sendKey = new SendKey(getKeyPort);
		sendKey.start();
		// Cryptography: end

		scommesse = new HashMap<Integer, Scommessa>();
		limite = deadline;
		this.port = port;
		accepter = new BetAccepter(port);
		accepter.start();
	}

	/**
	 * Thread per la gestione dell'invio della chiave pubblica
	 * 
	 * @author Raffaele Giordanelli
	 */
	private class SendKey extends Thread {

		private int sendKeyPort;
		private ServerSocket servSendKey;

		SendKey(int p) {
			try {
				// Creazione del ServerSocket sulla porta sendKeyPort
				sendKeyPort = p;
				servSendKey = new ServerSocket(sendKeyPort);

				// inizializza un generatore di coppie di chiavi usando RSA
				KeyPairGenerator kpg = KeyPairGenerator.getInstance("RSA");
				kpg.initialize(1024);
				// genera la coppia
				kp = kpg.generateKeyPair();

				// Inizializzo un cifrario che usa come algoritmo RSA, come
				// modalità ECB e come padding PKCS1
				decipher = Cipher.getInstance("RSA/ECB/PKCS1Padding");
				// Lo inizializzo dicendo modalità di codifica e chiave privata
				// da usare
				decipher.init(Cipher.DECRYPT_MODE, kp.getPrivate());

				// Inizializzo un cifrario che usa come algoritmo RSA, come
				// modalità ECB e come padding PKCS1
				cipher = Cipher.getInstance("RSA/ECB/PKCS1Padding");
				// Lo inizializzo dicendo modalità di codifica e chiave privata
				// da usare
				cipher.init(Cipher.ENCRYPT_MODE, kp.getPrivate());

			} catch (Exception e) {
				System.err.println(e);
			}
		}

		@Override
		public void run() {
			try {
				while (true) {
					Socket s = servSendKey.accept();

					InputStream in = s.getInputStream();
					in.read();

					OutputStream out = s.getOutputStream();
					out.write(kp.getPublic().getEncoded());
					out.close();
				}
			} catch (IOException ioe) {
				System.err.println(ioe);
			}
		}
	}

	/**
	 * Thread per gestire le scommesse ricevute in tempo utile
	 * 
	 * @author Raffaele Giordanelli
	 * 
	 */
	class BetAccepter extends Thread {

		private int port;
		private ServerSocket serv;
		private boolean accept;

		BetAccepter(int p) {
			try {
				port = p;
				serv = new ServerSocket(port);
				accept = true;
			} catch (IOException ioe) {
				System.err.println(ioe);
			}
		}

		@Override
		public void run() {

			while (accept) {
				try {
					Calendar now = Calendar.getInstance();
					serv.setSoTimeout((int) (limite.getTimeInMillis() - now.getTimeInMillis()));
					Socket k = serv.accept();
					InputStream in = k.getInputStream();
					PrintWriter out = new PrintWriter(k.getOutputStream(), true);

					// Cryptography: begin
					// Salvo la scomemssa inviata al server
					byte[] scommessaBytes = new byte[1000];
					int numByte = in.read(scommessaBytes);
					scommessaBytes = Arrays.copyOf(scommessaBytes, numByte);
					byte[] scommessaDecriptata = decipher.doFinal(scommessaBytes);

					// Cryptography: begin

					String line = new String(scommessaDecriptata);

					System.out.println("scommessa decifrata=" + line);

					int pos = line.indexOf(" ");
					int numCavallo = Integer.parseInt(line.substring(0, pos));
					long puntata = Long.parseLong(line.substring(pos + 1));
					InetAddress ip = k.getInetAddress();

					Scommessa s = new Scommessa(numCavallo, puntata, ip);
					int key = s.getID();
					scommesse.put(new Integer(key), s);

					// manda ok
					out.println("Scommessa accettata.");
					out.close();
					k.close();
					System.out.println("Ricevuta scommessa " + ip + " " + numCavallo + " " + puntata);

				} catch (SocketTimeoutException ste) {
					accept = false;
					System.out.println("Tempo a disposizione per le scommesse terminato");
				}// catch
				catch (IOException ioe) {
					System.err.println(ioe);
				} catch (Exception e) {
					System.err.println(e);
				}
			}// while
			try {
				serv.close();
			} catch (IOException ioe) {
				System.out.println(ioe);
			}
		}// run
	}// BetAccepter

	/**
	 * Thread per gestire le scommesse ricevute fuori tempo massimo
	 * 
	 * @author Raffaele Giordanelli
	 * 
	 */
	class BetDenier extends Thread {

		private int port;
		private ServerSocket serv;
		private boolean closed;

		BetDenier(int p) {
			try {
				port = p;
				serv = new ServerSocket(port);
				closed = true;
			} catch (IOException ioe) {
				System.err.println(ioe);
			}
		}

		public void reset() {
			closed = false;
		}

		public void run() {
			try {
				while (closed) {
					Socket k = serv.accept();
					PrintWriter out = new PrintWriter(k.getOutputStream(), true);
					out.println("Scommesse chiuse");
					out.close();
					k.close();
					System.out.println("Rifiutata scommessa");
				}// while
				serv.close();
			} catch (IOException ioe) {
				System.err.println(ioe);
			}
		}// run
	}// BetDenier

	/**
	 * Attende fino al termine del thread BetAccepter. Quando il metodo termina,
	 * le scommesse sono automaticamente chiuse.
	 */
	public void attendiChiusuraScommesse() {
		try {
			accepter.join();
		} catch (InterruptedException ie) {
			System.err.println(ie);
		}
	}

	/**
	 * Avvia il thread BetDenier che rifiuterà le scommesse in arrivo.
	 */
	public void chiusuraScommesse() {
		denier = new BetDenier(port);
		denier.start();
	}

	/**
	 * Fa terminare anche il thread BetDenier.
	 */
	public void resetServer() {
		denier.reset();
	}

	/**
	 * @return la lista delle scommesse vincenti, ovvero con puntata sul cavallo
	 *         passato come parametro
	 */
	public LinkedList<Scommessa> controllaScommesse(int cavalloVincente) {

		LinkedList<Scommessa> elenco = new LinkedList<Scommessa>();
		for (int i = 0; i < scommesse.size(); i++) {
			Scommessa s = scommesse.get(new Integer(i));
			if (s.equals(new Scommessa(cavalloVincente, 0, null)))
				elenco.addLast(s);
		}
		return elenco;
	}

	/**
	 * Invia a tutti i client in multicast l'elenco dei vincitori
	 */
	public void comunicaVincitori(LinkedList<Scommessa> vincitori, InetAddress ind, int port) {

		ListIterator<Scommessa> it = vincitori.listIterator();
		try {
			MulticastSocket socket = new MulticastSocket();
			byte[] buf = new byte[256];
			String m = "";
			while (it.hasNext()) {
				Scommessa s = it.next();
				m += s.getCavallo() + " " + (s.getPuntata() * 12) + "\n";
			}

			// Cryptography: begin
			// cripta con la chive privata i dati per inviarli allo
			// scommettitore
			buf = cipher.doFinal(m.getBytes());
			// Cryptography: end

			DatagramPacket pk = new DatagramPacket(buf, buf.length, ind, port);
			socket.send(pk);
		} catch (IOException ioe) {
			System.err.println(ioe);
		} catch (Exception e) {
			System.err.println(e);
		}
	}

	public static void main(String[] args) {

		int serverPort = 8001;
		int clientPort = 8002;

		// Cryptography: begin
		int getKeyServerPort = 8443;
		// Cryptography: end

		try {
			InetAddress multiAddress = InetAddress.getByName("230.0.0.1");

			Calendar deadline = Calendar.getInstance();
			deadline.add(Calendar.SECOND, 10);

			BetServer server = new BetServer(serverPort, getKeyServerPort, deadline);
			server.attendiChiusuraScommesse();
			server.chiusuraScommesse();
			int vincente = ((int) (Math.random() * 12)) + 1;
			System.out.println("E' risultato vincitore il cavallo " + vincente);
			LinkedList<Scommessa> elenco = server.controllaScommesse(vincente);
			server.comunicaVincitori(elenco, multiAddress, clientPort);
			System.out.println("Elenco vincitori:");
			System.out.println(elenco);

			Thread.sleep(150000);
			server.resetServer();
		} catch (InterruptedException ie) {
			System.err.println(ie);
		} catch (UnknownHostException uhe) {
			System.err.println(uhe);
		}

	}// main

}// BetServer

